//
// The calculations here form the heart of the fit program.
// Upon being fed the data (x,y,s) and a set of parameters
// (b), along with the form of the function that is attempted
// to be fit to the data f(x,b) (as described by the
// user-supplied function func.cc), this calculate() function will
// execute the iterative strategy described by Donald
// Marquardt so as to reduce phi (the sum of the squares of
// the residuals) to its minimum value.  When phi is
// minimized, the values of the parameters will be optimum,
// resulting in a "least-squares" fit of f(x,b) to the data.
//
// The steps:
//
// (a) Start by showing the list of parameters, the initial
//     (estimated) parameter values and the initial value of phi
//     (derived from the initial parameter values), which we save
//     in phi_old.
//
// (b) Use get_g_a() to find the npar components of g, the
//     (negative) gradient vector (which points downward in a
//     direction perpendicular to the contour lines of phi),
//     along with the components of a, the curvature matrix of
//     the npar-dimensional phi surface. Save the (unscaled) a-
//     matrix components in a_save, as they will be needed later
//     in step (j).  The components of g and a will depend on
//     the current values of the parameters, i.e., on the
//     components of the b-vector.
//
// (c) Scale both g and a, after which g and a will be the scaled
//     vector and matrix. a will be symmetric, with values of the
//     diagonal elements equal to 1.00.
//
// (d) Copy (the scaled) g to an npar-row, 1-column matrix g1.
//     This will be needed by gaussj() (the Gauss-Jordan
//     elimination routine), used in step (f).
//
// (e) Add the value of lambda to each diagonal element of a
//     to create the matrix a + lambda*I.
//
// (f) Use gaussj() to solve (a + lambda*I)*db = g1 to obtain the
//     (scaled) column vector of corrections db to the parameter
//     vector b. db will point toward the minimum of the phi
//     surface.
//
// (g) Unscale the db components and add them to the b
//     components, so as to obtain a new b-vector that we hope
//     will reduce phi, i.e., we hope that phi(b+db) < phi(b).
//
// (h) If phi(b+db) >= phi(b), multiply lambda by 5, and
//     return to step (b).
//
// (i) If phi(b+db) < phi(b), and test for convergence, using the
//     "epsilon test", which judges whether all components of the
//     scaled db are small enough to stop. If the epsilon test
//     fails (the db components are not small enough), divide
//     lambda by 5, and return to step (b) to continue iterating.
//
// (j) If the epsilon test passes, do one additional iteration
//     with lambda = 0.0, and show the final results, complete
//     with standard errors for each of the parameters, including
//     the "student" t-factors.
//

#include "fit.h"

void calculate(dvec x, dvec y, dvec s, dvec& b, misc m, opts o)
{
  uns npar = m.npar = b.size();
  uns npts;                                        // number of data points
  uns nu;                                             // degrees of freedom
  int j;
  int fw;                                                    // field width
  m.maxiter = 20;                              // maximum no. of iterations
  fail_pass k;                                      // k=0: fail, k=1: pass
  int n;                                               // iteration counter
  string ss;                                    // text for show() function
  string fs;                                   // form of func for graphing
  osstream oss;
  dvec p;                                   // partial derivatives df/db[j]
  dbl phi, phi_old;                                         // value of phi
  dbl gam;      // gamma = angle in degs between (negative) gradient and db
  dbl lam = m.lambda = 1.0e-2;                        // init val of lambda
  dbl t;                                                // student t-factor
  dvec g(npar);                                 // negative gradient vector
  dvec g_save(npar);                               // for confidence limits
  dvec db(npar);                // transform scaled gcol soln to row vector
  dvec sf(npar);                                     // scale factor vector
  dmat a(npar, dvec(npar));                             // curvature matrix
  dmat a_save(npar);                // save unscaled a for future unscaling
  dmat gcol(npar, dvec(1));                           // g in column format
  cout.precision(4);         // sig figs in displays of intermediate output

  dbl f = func(x[0],b,p,fs,m);          // need func to get names of params

  // make a time/date string
  time_t now = time(NULL);
  struct tm* info = localtime(&now);
  const string fmt = "  Results at %I:%M:%S %p on %A, %B %d, %Y:";
  string td;                                            // time/date string
  gstrftime(td, fmt, info);
  cout << "\n" << td << "\n\n";
  fw = 26;
  cout << right;
  cout << setw(fw) << "datafile name = " << m.datafile << "\n";
  npts = m.npts = x.size();
  nu = npts - npar;

  cout << setw(fw) << "number of data points = " << m.npts << "\n";
  cout << setw(fw) << "degrees of freedom = " << nu << "\n";
  cout << setw(fw) << "independent sigmas?  ";
  string sigs = o.sigmas ? "yes" : "no";
  cout << sigs << "\n\n";
  if (!m.ssize)
    cout << "  You have not supplied sigmas in your\n"
         << "  datafile, so I'm assuming you want to\n"
         << "  give each of your points equal weight.\n\n";

  cout << "  parameter names and their initial values:\n\n";

  int maxlen = 0;
  for (j=0;j<npar;j++)                 // get max length of name[j] strings
    maxlen = (m.name[j].length() > maxlen) ? m.name[j].length() : maxlen;
  fw = maxlen + 2;                               // field width for display
  cout << right << showpoint;
  for (j=0;j<npar;j++)
    cout << setw(fw) << m.name[j] << " = " << b[j] << "\n";

  phi = get_phi(x,y,s,b,m);     // gets value of phi with existing b-vector
  cout << "\n  initial phi = " << showpoint << phi << "\n";
  phi_old = phi;                                                // save phi

  for (n=1;n<=m.maxiter;n++) {               // limit to maxiter iterations
    cout << "\n ===========================\n";

    get_g_a(g,a,x,y,s,b,m);              // gets current (negative) g and a
                           // these depend on current b
    a_save = a;                        // save a for later use in unscaling

    scale(g,a,sf,npar);                          // destroys unscaled g & a
    gcol = copy(g,1);         // gcol is g as column vector, used in gaussj

    cout << "  iteration #" << n << ", with lambda = " << lam << ":\n";

    for (j=0;j<npar;j++)
      a[j][j] += lam;                       // add lam to diags of scaled a

    gaussj(a,gcol);                // after solving, gcol will be scaled db

    db = copy(gcol,1);                            // db is gcol in row form
    if (o.verbose) {
      cout.precision(3);
      ss = "\n  scaled correction vector db:";
      show(db,ss);
      gam = gamma(db,g,npar);
      cout << "  gamma = " << gam << " degrees\n";
      cout.precision(4);
    }

    for (j=0;j<npar;j++)                                      // unscale db
      db[j] /= sqrt(a_save[j][j]);

    for (j=0;j<npar;j++)
      b[j] += db[j];                                          // new params

    phi = get_phi(x,y,s,b,m);   // gets value of phi with existing b-vector
    cout << "  phi = " << phi << "\n";
    if (phi < phi_old) {
      if (o.verbose)
        cout << "  phi decreases, so we decrease lambda\n";
      lam /= 5.0;
    }
    else {
      if (o.verbose)
        cout << "  phi increases, so we increase lambda\n";
      lam *= 5.0;
    }
    phi_old = phi;                                          // save new phi
    ss = "\n  parameter values:";
    show(b,ss);

    k = eps_test(b,db,sf,npar);
    if (k == 1) {        // passes eps_test, so set lam=0 and do final iter
      cout << "\n  The epsilon test for convergence is passed,\n"
           << "  so we iterate once more with lambda = 0.0\n";
      cout << "\n ===========================\n";
      get_g_a(g,a,x,y,s,b,m);
      a_save = a;
      g_save = g;                  // used in calculating confidence limits
      scale(g,a,sf,npar);
      gcol = copy(g,1);
      gaussj(a,gcol);
      scale(g,a,sf,npar);
      db = copy(gcol,1);
      for (j=0;j<npar;j++)                                    // unscale db
        db[j] /= sqrt(a_save[j][j]);
      for (j=0;j<npar;j++)
        b[j] += db[j];                                        // new params
      break;
    }
  }
  // now print the final results
  gcol = copy(g_save,1);
  gaussj(a_save,gcol);                   // inverse of a_save now in a_save
  dvec serr(npar);                                  // vector of std errors
  cdbl conf = 0.68269;                              // one-sigma conf level
  t = student(nu,conf);
  phi = get_phi(x,y,s,b,m);                           // final value of phi
  for (j=0;j<npar;j++) {
    if (o.sigmas)                                  // sigmas option invoked
      serr[j] = t*sqrt(a_save[j][j]);
    else                                        // no sigmas option invoked
      serr[j] = t*sqrt(phi*a_save[j][j]/nu);
  }
  cout.precision(5);
  cout << "\n  Here are the final parameter values, with\n"
       << "  one-sigma (68.27%) standard errors:\n\n";
  cout << right << showpoint;
  fw = maxlen + 2;
  for (j=0;j<npar;j++)
    cout << setw(fw) << m.name[j] << " = "
         << b[j] << " +/- " << serr[j] << "\n";
  ss = "\n  And here is the correlation matrix:\n";
  show(a,ss);
  cout << "\n";
  m.b = b;                     // save current params for use in mk_graph()
  if (o.sigmas && m.ssize) { // chi-square good only for independent sigmas
    dbl alpha = gammq(0.5*nu, 0.5*phi);
    oss.precision(4);
    cout.precision(4);
    fw = 34;
    oss.clear();                                     // ensure oss is empty
    oss << phi/nu;
    ss = oss.str();
    cout << right << showpoint
         << setw(fw) << "  chi-square = " << phi << "\n"
         << setw(fw) << "  chi-square per deg of freedom = " << ss << "\n"
         << setw(fw) << "  alpha = " << alpha << "\n"
         << "\n"
            "   alpha is the probability that a larger value\n"
            "   of chi-square would be obtained for an equivalent\n"
            "   data set, assuming the data to be drawn from a\n"
            "   normal distribution.\n\n";
  }
  else if (!o.sigmas && m.ssize) {       // sigmas are present but relative
    cout << "   No meaningful values of either chi-square or its\n"
            "   associated probability alpha are available without\n"
            "   independent sigmas.  I'm assuming your sigmas are\n"
            "   relative, and not independently calculated.\n"
            "\n";
  }
  else {                           // no sigmas are supplied (m.ssize == 0)
    cout << "   No meaningful values of either chi-square or its\n"
            "   associated probability alpha are available without\n"
            "   independent sigmas.  I see no sigmas in your datafile\n"
            "   so I'm assuming you intend to give equal weight to\n"
            "   each point.\n"
            "\n";
  }
  cerr << "  Would you like to see a graph of your fit? ([yes] no) ";
  if (query_user()) { // "return" or any word with "y" or "Y" returns a "1"
    cout << "  OK, I'll make a graph...\n\n";
    mk_graph(m,o);
  }
  else
    cout << "  OK, no graph this time...\n\n";
}
