//
// All of these are from numerical recipes in C 3/e,
// published in 2007.  In NR 3/e the incomplete gamma parts
// are presented as components of an object Gamma, but here we
// just present them as separate parts.  There is no need to
// incorporate nr3.h, as these routines are slightly modified
// to avoid its need.  September 30, 2014
//

#include "fit.h"

// these are for NR 3/e code
cdbl EPS = numeric_limits<dbl>::epsilon();
cdbl FPMIN = numeric_limits<dbl>::min()/EPS;
dbl gln;

//
// this next version from NR 3/e, page 258.
// returns ln(n!)
//
dbl factln(cint n)
{
  static cint NTOP=2000;
  static dvec a(NTOP);
  static bool init=true;
  if (init) {
    init = false;
    for (int i=0;i<NTOP;i++) a[i] = gammln(i+1.0);
  }
  if (n < 0) throw("negative arg in factln");
  if (n < NTOP) return a[n];
  return gammln(n+1.0);                            // out of range of table
}

//
// gammln from NR 3/e, page 257:
// returns value of ln(GAMMA(xx) for xx > 0
//
dbl gammln(cdbl xx)
{
  int j;
  dbl x, tmp, y, ser;
  static cdbl cof[14]={57.1562356658629235,-59.5979603554754912,
  14.1360979747417471,-0.491913816097620199,.339946499848118887e-4,
  .465236289270485756e-4,-.983744753048795646e-4,.158088703224912494e-3,
  -.210264441724104883e-3,.217439618115212643e-3,-.164318106536763890e-3,
  .844182239838527433e-4,-.261908384015814087e-4,.368991826595316234e-5};
  if (xx <= 0) throw("bad arg in gammln");
  y=x=xx;
  tmp = x+5.24218750000000000;                          // Rational 671/128
  tmp = (x+0.5)*log(tmp)-tmp;
  ser = 0.999999999999997092;
  for (j=0;j<14;j++) ser += cof[j]/++y;
  return tmp+log(2.5066282746310005*ser/x);
}

//
// Here is gammq from NR 3/e, page 261, which returns the
// incomplete gamma function Q(a,x) = 1 - P(a,x)
//

dbl gammq(cdbl a, cdbl x)
{
  if (x < 0.0 || a <= 0.0) throw("bad args in gammq");
  if (x == 0.0) return 1.0;
  else if (x < a+1.0) return 1.0-gser(a,x);           // use the series rep
  else return gcf(a,x);                   // use the continued fraction rep
}

//
// Here is gser from NR 3/e, page 261
// Returns incomplete gamma function P(a,x) evaluated by its
// series representation, and sets gamma(a) as gln.
//
dbl gser(cdbl a, cdbl x)
{
  dbl sum,del,ap;
  gln = gammln(a);
  ap = a;
  del = sum = 1.0/a;
  for (;;) {
    ++ap;
    del *= x/ap;
    sum += del;
    if (fabs(del) < fabs(sum)*EPS) {
      return sum*exp(-x+a*log(x)-gln);
    }
  }
}

//
// Here is gcf from NR 3/e, page 261.
// It returns incomplete gamma function Q(a,x) evaluated by
// its continued fraction representation
//
dbl gcf(cdbl a, cdbl x)
{
  int i;
  dbl an,b,c,d,del,h;
  gln=gammln(a);
  b=x+1.0-a;
  c=1.0/FPMIN;
  d=1.0/b;
  h=d;
  for (i=0;;i++) {
    an = -i*(i-a);
    b += 2.0;
    d = an*d+b;
    if (fabs(d) < FPMIN) d = FPMIN;
    c = b+an/c;
    if (fabs(c) < FPMIN) c = FPMIN;
    d = 1.0/d;
    del = d*c;
    h *= del;
    if (fabs(del-1.0) <= EPS) break;
  }
  return exp(-x+a*log(x)-gln)*h;
}
