// October 21, 2014
//
// This is "func.cc", the single C++ function that must be
// supplied by the user in the "fit" program.
//
// It contains the form of the function to be fit to the data,
// along with its partial derivatives p[j] with respect to the
// parameters b[j] ("p" for "partial").
//
// This example of func.cc illustrates what would be suitable
// for fitting data described by a single Gaussian lineshape.
// It works for the "gaussdata" testfile supplied in this
// package.
//
// Lines marked with "// $$$$" are the only ones you need to
// change if you want to use a different function to describe
// your data.
//
// See fit -h for help on how to make this file.
//

#include "fit.h"

dbl func(dbl x, dvec b, dvec& p, string& fs, misc& m)
{

// f is the value of the function to be fit to the data. f will
// depend on the values of the current parameters, and the value
// of f is returned when func() is called. The b[k] are the
// parameters, supplied in the argument of func() The p[k] are
// the partial derivatives df/db[k] ("p" for "partial")

  dbl f;

//  f = b[0]*exp(-(pow(x-b[1],2)/(2.0*pow(b[2],2)))) + b[3];
//  The above works, but I think it's less confusing to avoid
//  using pow(stuff,n) when n is only 2 or 3:
  f = b[0]*exp(-(x-b[1])*(x-b[1])/(2.0*b[2]*b[2])) + b[3];          // $$$$

// We make up the string fs for f(x,b), for use in the Python
// script for the graph. This is tricky, since we have to
// interpolate the values of the b[k] into the string. If, when
// you try to display the graph of your fit, you get an
// inexplicable error message from Python, it is likely that you
// have not made the fs string correctly. Its format will be
// displayed on the screen when you request the graph, so you can
// check to make sure it is correct. func(x,b) will be called
// with the final values of the b[k], just prior to making the
// graph. Here's how we make the string fs:

  osstream oss;                                      // output stringstream
  oss.clear();
  oss.precision(4);
  oss << b[0] << "*exp(-(x-"                                        // $$$$
      << b[1] << ")**2/(2.0*"                                       // $$$$
      << b[2] << "**2)) + "                                         // $$$$
      << b[3] << "\n";                                              // $$$$
  //
  // fs: the string form of our function, returned as the fourth
  // argument of func().
  //
  fs = oss.str();                              // convert oss into a string

// Note that for the Python script (but not for this C++),
// stuff**2 is just the square of stuff.
//
// Uncomment the next "cout" line to check to make sure it
// displays correctly on your screen.
//
// cout << "  fs = " << fs << "\n";
//
// The number of parameters is b.size(), the size (number of
// elements) of the parameter vector:

  int npar = b.size();
  m.npar = npar;                       // store npar in the misc structure.

// The names of the parameters.  They should not be too long.
  svec name(npar);                                   // declare name vector
  name[0] = "amplitude = b[0]";                                     // $$$$
  name[1] = "mean = b[1]";                                          // $$$$
  name[2] = "width = b[2]";                                         // $$$$
  name[3] = "base = b[3]";                                          // $$$$
  m.name = name;                 // store name vector in the misc structure

// The values of p[k] = df/db[k] (partial derivatives). Be very
// careful in working these out. Errors in this code are the main
// cause of a non-functioning fit program.

  p.resize(m.npar);
  p[0] = (f-b[3])/b[0];                                             // $$$$
  p[1] = (f-b[3])*(x-b[1])/(b[2]*b[2]);                             // $$$$
  p[2] = (f-b[3])*(x-b[1])*(x-b[1])/(b[2]*b[2]*b[2]);               // $$$$
  p[3] = 1.0;                                                       // $$$$

  return f;    // The value of the function returned when func() is called.
}
